/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.1)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2005 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* CVS $Id: l_comple.cpp,v 1.7 2005/09/01 08:55:49 cxsc Exp $ */

#include "l_comple.hpp"  // corresponding header file
#include "l_interv.hpp"

namespace cxsc {

// ---------------- Unary Operators -----------------------------------------
l_complex operator-(const l_complex& x)
{
    return l_complex(-x.re, -x.im);
}

l_complex operator+(const l_complex& x)
{
    return x;
}

// ----------------- cdotprecision +/- l_complex ----------------------------

cdotprecision operator+(const l_complex& lc, const cdotprecision& cd) throw() 
{ 
   return _cdotprecision(lc) + cd; 
}

cdotprecision operator+(const cdotprecision& cd, const l_complex& lc) throw() 
{ 
   return _cdotprecision(lc) + cd; 
}   

cdotprecision operator-(const l_complex& lc, const cdotprecision& cd) throw() 
{ 
   return _cdotprecision(lc) - cd; 
}

cdotprecision operator-(const cdotprecision& cd, const l_complex& lc) throw() 
{ 
   return cd - _cdotprecision(lc); 
}

// ------------------ dotprecision +/- l_complex ----------------------------

cdotprecision operator+(const l_complex& lc, const dotprecision& cd) throw() 
{ 
   return _cdotprecision(lc) + cd; 
}

cdotprecision operator+(const dotprecision& cd, const l_complex& lc) throw() 
{ 
   return _cdotprecision(lc) + cd; 
}   

cdotprecision operator-(const l_complex& lc, const dotprecision& cd) throw() 
{ 
   return _cdotprecision(lc) - cd; 
}

cdotprecision operator-(const dotprecision& cd, const l_complex& lc) throw() 
{ 
   return cd - _cdotprecision(lc); 
}

// ---------------- l_complex * l_complex -----------------------------------
l_complex operator * (const l_complex& a, const l_complex& b)
    throw()
{
    l_real r, i; 
    dotakku[0] = 0;
    accumulate(dotakku[0],a.re,b.re);
    accumulate(dotakku[0],-a.im,b.im);
    r = dotakku[0];
    dotakku[0] = 0;
    accumulate(dotakku[0],a.im,b.re);
    accumulate(dotakku[0],a.re,b.im);
    i = dotakku[0];
    return( l_complex(r,i) );            
}

// ------------------ l_complex * complex -----------------------------------
l_complex operator * (const l_complex& a, const complex& b)
    throw()
{
    l_real r, i; 
    dotakku[0] = 0;
    accumulate(dotakku[0],a.re,Re(b));
    accumulate(dotakku[0],-a.im,Im(b));
    r = dotakku[0];  // r is real-part
    dotakku[0] = 0;
    accumulate(dotakku[0],a.im,Re(b));
    accumulate(dotakku[0],a.re,Im(b));
    i = dotakku[0];  // i is imaginary part
    return( l_complex(r,i) );
}

l_complex operator * ( const complex& b, const l_complex& a )
    throw()
{
    l_real r, i; 
    dotakku[0] = 0;
    accumulate(dotakku[0],a.re,Re(b));
    accumulate(dotakku[0],-a.im,Im(b));
    r = dotakku[0];  // r is real part
    dotakku[0] = 0;
    accumulate(dotakku[0],a.im,Re(b));
    accumulate(dotakku[0],a.re,Im(b));
    i = dotakku[0];  // i is imaginary part
    return( l_complex(r,i) );
}

// ---------------- l_complex / l_complex -----------------------------------

static const int maxexpo1 = 1022, Min_Exp_ = 1074,
                 minexpom = -50,  minexpo1 = -1124; 

void product(l_real a, l_real b, l_real c, l_real d, 
             int& overfl, l_interval& res)
// a,b,c,d and res must have the same precision!
// Input data: a,b,c,d;  Output data: overfl, res;
// In case of overfl=0 the interval res is an inclusion of a*b + c*d;
// overfl=1 (overflow) signalizes that res must be multiplied with 2^1074
// to be an inclusion of a*b + c*d;
// overfl=-1 (underflow) signalizes that res must be multiplied with 
// 2^-1074 to be an inclusion of a*b + c*d; 
// overfl=-2 (underflow) signalizes that res must be multiplied twice with 
// 2^-1074 to be an inclusion of a*b + c*d;

{
    int exa,exb,exc,exd; // Exponents of a,b,c,d;
    interval z;
    l_real tmp;
    l_interval li1,li2;
    dotprecision dot;
    idotprecision idot;
    int base,skf1,skf2;

    overfl = 0;  // Normal case, no overflow or undeflow;
    res = 0;  dot = 0;

    exa = expo(a[1]);
    exb = expo(b[1]);   
    exc = expo(c[1]);
    exd = expo(d[1]);

    if ( sign(a[1])==0 || sign(b[1])==0 )    // a*b == 0; 
       if ( sign(c[1])==0 || sign(d[1])==0 ) // c*d == 0; 
	   ; // No operation necessary,  res = 0;
       else
       { //  a*b == 0;  c*d != 0;
	   if (exc+exd > maxexpo1) 
	   { // a*b == 0; overflow originated by c*d;
	       overfl = 1;
	       base = (exc+exd-Min_Exp_) / 2; // to get optimal accuracy
	       skf1 = base - exc;        // Scaling exponent for c
	       skf2 = -skf1 - Min_Exp_;  // Scaling exponent for d
               times2pown(c,z,skf1);     // Scaling c with skf1
               l_realz2l_interval(c,z,li1); // li1 is an inclusion of c+z 
               times2pown(d,z,skf2);     // Scaling d with skf2
	       l_realz2l_interval(d,z,li2); // li2 is an inclusion of d+z
	       res = li1*li2; 
	   } else
	   {
	       if ( exc+exd < minexpom )
	       { // a*b == 0; too small product values originated by c*d
		   if (exc+exd<minexpo1)
	           {
		       overfl = -2;
		       times2pown(c,Min_Exp_); // first scaling  (exact)
		       times2pown(d,Min_Exp_); // second scaling (exact)
	           } else
	           { 
	               overfl = -1;
	               if (exc<exd) times2pown(c,Min_Exp_); // exact scaling!
	               else times2pown(d,Min_Exp_); // exact scaling!
	           }
	       } 
	       accumulate(dot,c,d);
	       res = l_interval(dot);
	   }  
       } else // now a*b != 0
       if ( sign(c[1])==0 || sign(d[1])==0 ) 
       { // a*b !=0,  c*d == 0:
	   if (exa+exb > maxexpo1) 
	   { // overflow originated by a*b;
	       overfl = 1;
	       base = (exa+exb-Min_Exp_) / 2;
	       skf1 = base - exa;        // Scaling for a
	       skf2 = -skf1 - Min_Exp_;  // Scaling for b
               times2pown(a,z,skf1);
               l_realz2l_interval(a,z,li1);
               times2pown(b,z,skf2);
	       l_realz2l_interval(b,z,li2);
	       res = li1*li2; 
	   } else
	   {
	       if ( exa+exb < minexpom )
	       {   // underflow originated by a*b
		   if (exa+exb<minexpo1)
	           {
		       overfl = -2;
                       times2pown(a,Min_Exp_);
                       times2pown(b,Min_Exp_);
	           } else
	           {
	               overfl = -1;
	               if (exa<exb) times2pown(a,Min_Exp_); // exact scaling!
	               else times2pown(b,Min_Exp_);         // exact scaling!
	           }
	       } 
	       accumulate(dot,a,b);
	       res = l_interval(dot);
	   }
       } else  // a,b,c,d != 0:
       {
	   if (exc+exd > exa+exb)
	   {   // interchange a and c:
	       tmp = a;  a = c;  c = tmp;
	       base = exa;  exa = exc;  exc = base;
               // interchange b and d:
	       tmp = b;  b = d;  d = tmp;
	       base = exb;  exb = exd;  exd = base;
	   }  // From now on it holds:  exa+exb >= exc+exd;
	   if (exa+exb > maxexpo1)
	   { // overflow by a*b;  c*d != 0;
	       overfl = 1;  idot = 0;
	       base = (exa+exb-Min_Exp_) / 2;
	       skf1 = base - exa;        // Scaling exponent for a
	       skf2 = -skf1 - Min_Exp_;  // Scaling exponent for b
               times2pown(a,z,skf1);  // Scaling a
               l_realz2l_interval(a,z,li1);  // a+z == li1
               times2pown(b,z,skf2);  // Scaling b
	       l_realz2l_interval(b,z,li2);  // b+z == li2
               accumulate(idot,li1,li2);
	       base = (exc+exd-Min_Exp_) / 2;
	       skf1 = base - exc;        // Scaling exponent for c
	       skf2 = -skf1 - Min_Exp_;  // Scaling exponent for d
               times2pown(c,z,skf1);  // Scaling c
               l_realz2l_interval(c,z,li1);  // c+z == li1
               times2pown(d,z,skf2);  // Scaling d
	       l_realz2l_interval(d,z,li2);  // d+z == li2
	       accumulate(idot,li1,li2);
	       res = l_interval(idot);
	   } else
	   {
	       if (exa+exb < minexpom)
	       {   
		   if (exa+exb<minexpo1)
		   {
		       overfl = -2; // exact double scaling:
		       times2pown(a,Min_Exp_); 
		       times2pown(b,Min_Exp_);
		       times2pown(c,Min_Exp_);
		       times2pown(d,Min_Exp_);
		   } else
	           { 
		       overfl = -1; // exact scaling:
	               if (exa<exb) times2pown(a,Min_Exp_); // exact scaling!
	               else times2pown(b,Min_Exp_);         // exact scaling!
	               if (exc<exd) times2pown(c,Min_Exp_); // exact scaling!
	               else times2pown(d,Min_Exp_);         // exact scaling!
		   }
	       } 
	       accumulate(dot,a,b);
	       accumulate(dot,c,d);
	       res = l_interval(dot);
	   }
       } // a,b,c,d != 0
} // end product

l_real quotient(l_interval z, l_interval n, int round, 
		int zoverfl, int noverfl)
// z is an inclusion of a numerator.
// n is an inclusion of a denominator.
// quotient(...) calculates with q1 an approximation of z/n
// using staggered arithmetic.
// Rounding with round (-1,0,+1) is considered.
// zoverfl and noverfl are considered by scaling back with 2^1074 or 2^-1074
// n1+n2 > 0 is assumed.
{   
    l_real q1;
    int vorz, anz_scale, sc, ex=0; 
    l_interval res;

    if ( zero_(z) && !zero_(n) ) { q1=0;  return q1; };

    vorz = sign(z[1]);

    if ( zoverfl-noverfl<=-2 ) // result in underflow
    {
	switch(round)
	{
	    case RND_DOWN:
		if (vorz >= 0) q1 = 0.0;
		else q1 = -minreal;
		break;
	    case RND_NEXT:
		q1 = 0.0;
		break;
	    case RND_UP:
		if (vorz <= 0) q1 = 0.0;
		else q1 = minreal;
		break;
	} // switch
    } else if ( zoverfl-noverfl>=2 )  // result in overflow
    {
	q1[1] = comp(0.5,1050); // produces: segmentation fault
    } else
    {
	res = z / n;
	anz_scale = zoverfl - noverfl;
        // Back-scaling in two steps is necessary:
	if (anz_scale > 0) { sc = 1023; ex = 51; }
	else if (anz_scale < 0) { sc = -1023; ex = -51; }
	     else sc = 0;
	if (sc) times2pown(res,sc);  // Back-scaling in two
	if (ex) times2pown(res,ex);  // steps now realized
        // Rounding:
	switch (round)
	{
	    case RND_DOWN:
		q1 = Inf(res);  break;
	    case RND_NEXT:
		q1 = mid(res);  break;
	    case RND_UP:
		q1 = Sup(res);  break;
	} // switch
    }
    return q1;
} // quotient

l_complex _c_division(l_complex a, l_complex b, int round)
{
    int zoverfl, noverfl;
    l_interval zl, nl;
    l_complex tmp;

    product(Re(b),Re(b),Im(b),Im(b),noverfl,nl);
    product(Re(a),Re(b),Im(a),Im(b),zoverfl,zl);
    SetRe(tmp, quotient(zl,nl,round,zoverfl,noverfl));
    product(Im(a),Re(b),-Re(a),Im(b),zoverfl,zl);
    SetIm(tmp, quotient(zl,nl,round,zoverfl,noverfl));
    return tmp;
} // _c_division

l_complex divn (const l_complex & a, const l_complex & b)  
{  
   return _c_division(a,b,RND_NEXT);
}

l_complex divd (const l_complex & a, const l_complex & b) 
{  
   return _c_division(a,b,RND_DOWN);
}

l_complex divu (const l_complex & a, const l_complex & b)  
{  
   return _c_division(a,b,RND_UP);
}

l_complex operator / (const l_complex &a, const l_complex &b) throw()
{
   return divn(a,b);
}

int StagPrec(const l_complex& lc) throw()
{
    return StagPrec(lc.re);
}      

void accumulate(cdotprecision& cd, const l_complex& lc1, 
                                   const l_complex& lc2) throw()
{
    accumulate(Re(cd),lc1.re,lc2.re); 
    accumulate(Re(cd),-lc1.im,lc2.im);
    accumulate(Im(cd),lc1.im,lc2.re);
    accumulate(Im(cd),lc1.re,lc2.im);
}

void accumulate(cdotprecision& cd, const l_complex& lc, 
                                   const complex& c) throw()
{
    accumulate(Re(cd),lc.re,Re(c)); 
    accumulate(Re(cd),-lc.im,Im(c));
    accumulate(Im(cd),lc.im,Re(c));
    accumulate(Im(cd),lc.re,Im(c));
}

void accumulate(cdotprecision& cd, const l_complex& lc, 
                                   const real& r) throw()
{
    accumulate(Re(cd),lc.re,r); 
    accumulate(Im(cd),lc.im,r);
}

void accumulate(cdotprecision& cd, const l_complex& lc, 
                                   const l_real& lr) throw()
{
    accumulate(Re(cd),lc.re,lr); 
    accumulate(Im(cd),lc.im,lr);
}

l_real abs2(const l_complex &a) throw()
{
   dotakku[4]=0;
   accumulate(dotakku[4],a.re,a.re);
   accumulate(dotakku[4],a.im,a.im);
   return l_real(dotakku[4]);
}

l_real abs(const l_complex& z) throw()
// Calculation of an approximation of |z|.
// In general the maximum precision is stagprec=19, predifined by the used
// sqrt-function declared in l_rmath.hpp.
// If the difference |exa-exb| of the exponents (base 2) is sufficient high,
// precision and accuracy can be choosen greater 19.
{
    return sqrtx2y2(Re(z),Im(z));
} // abs(...)


} // end namespace cxsc

















